#!/usr/bin/perl -w

# mailgraph -- a postfix statistics rrdtool frontend
# copyright (c) 2000, 2001 David Schweikert <dws@ee.ethz.ch>
# released under the GNU General Public License

use RRDs;
use Time::Local;

use strict;
use File::Tail;
use Getopt::Long;

my $VERSION = 0.14;

# config
my $rrdstep = 60;
my $xpoints = 540;
my $points_per_sample = 3;

# global variables
my $logfile;
my $rrd = "mailgraph.rrd";
my $year;
my $this_minute;
my %sum = ( sent => 0, received => 0, bounced => 0, rejected => 0 );
my $rrd_inited=0;
my $syslog_re = qr'^(...)[ ]{1,2}(\d+) (\d+):(\d+):(\d+) (\S+) (\S+?): (.*)'o;

# prototypes
sub process_line($);
sub syslogtime2time($$$$$);
sub event_sent($);
sub event_received($);
sub event_bounced($);
sub event_rejected($);
sub init_rrd($);
sub update($);

sub usage
{
	print "usage: mailgraph [*options*]\n\n";
	print "  -c, --cat          causes the logfile to be only read and not monitored\n";
	print "  -l, --logfile f    monitor logfile f instead of /var/log/syslog\n";
	print "  -h, --help         display this help and exit\n";
	print "  -v, --version      output version information and exit\n";

	exit;
}

sub main
{
	my %opt = ();
	GetOptions(\%opt, 'help|h', 'cat|c', 'logfile|l=s', 'version|v') or usage;
	usage if $opt{help};

	if($opt{version}) {
		print "mailgraph $VERSION by dws\@ee.ethz.ch\n";
		exit;
	}

	my $logfile = defined $opt{logfile} ? $opt{logfile} : '/var/log/syslog';

	if($opt{cat}) {
		open(FILE, "<$logfile") or die "can't open $logfile\n";
		while(<FILE>) {
			process_line($_);
		}
	}
	else {
		my $file = File::Tail->new(name=>$logfile, tail=>-1);
		my $line;
		while (defined($line=$file->read)) {
			process_line($line);
		}
	}
}

sub init_rrd($)
{
	my $m = shift;
	if(not -f $rrd) {
		my $rows = $xpoints/$points_per_sample;
		my $realrows = int($rows*1.1); # ensure that the full range is covered
		my $day_steps = int(3600*24 / ($rrdstep*$rows));
		# use multiples, otherwise rrdtool could choose the wrong RRA
		my $week_steps = $day_steps*7;
		my $month_steps = $week_steps*5;
		my $year_steps = $month_steps*12;
		RRDs::create($rrd, '--start', $m, '--step', $rrdstep,
				'DS:sent:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:recv:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:bounced:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:rejected:ABSOLUTE:'.($rrdstep*2).':0:U',
				"RRA:AVERAGE:0.5:$day_steps:$realrows",   # day
				"RRA:AVERAGE:0.5:$week_steps:$realrows",  # week
				"RRA:AVERAGE:0.5:$month_steps:$realrows", # month
				"RRA:AVERAGE:0.5:$year_steps:$realrows",  # year
				"RRA:MAX:0.5:$day_steps:$realrows",   # day
				"RRA:MAX:0.5:$week_steps:$realrows",  # week
				"RRA:MAX:0.5:$month_steps:$realrows", # month
				"RRA:MAX:0.5:$year_steps:$realrows",  # year
				);
		$this_minute = $m;
	}
	else {
		$this_minute = RRDs::last($rrd) + $rrdstep;
	}
	$rrd_inited=1;
}

sub syslogtime2time($$$$$)
{

	my ($mstr,$d,$H,$M,$S) = @_;
	my %months = (
		'Jan' => 0, 'Feb' =>  1, 'Mar' =>  2,
		'Apr' => 3, 'May' =>  4, 'Jun' =>  5,
		'Jul' => 6, 'Aug' =>  7, 'Sep' =>  8,
		'Oct' => 9, 'Nov' => 10, 'Dec' => 11 );
	my $m = $months{$mstr};
	if(not defined $m) {
		warn "Can't parse month '$1'\n";
		return;
	};

	$year = (localtime(time))[5] unless defined $year;
	my $time = timelocal($S,$M,$H,$d,$m,$year);
	if(defined $this_minute and $time < $this_minute) {
		# Dec 31 -> Jan  1 
		$year = (localtime(time))[5];
		$time = timelocal($S,$M,$H,$d,$m,$year);
	}
	return $time;
}

sub process_line($)
{
	my $line = shift;
	# parse syslog syntax
	$line =~ /$syslog_re/o or return;
	my ($m,$d,$H,$M,$S) = ($1,$2,$3,$4,$5);
	my $host = $6;
	my $prog = $7;
	my $str = $8;
	$prog =~ /^postfix\/(.*)/ or return; $prog = $1;
	my $time = syslogtime2time($m,$d,$H,$M,$S);
	$prog =~ s/\[\d+\]//; # strip process number
	if($prog eq 'smtp') {
		if($str =~ /\bstatus=sent\b/) {
			event_sent($time);
		}
		if($str =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'local') {
		if($str =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'smtpd') {
		if($str =~ /^[0-9A-F]+: client=/) {
			event_received($time);
		}
		if($str =~ /^reject: /) {
			event_rejected($time);
		}
	}
}

sub event_sent($)
{
	my $t = shift;
	update($t) and $sum{sent}++;
}

sub event_received($)
{
	my $t = shift;
	update($t) and $sum{received}++;
}

sub event_bounced($)
{
	my $t = shift;
	update($t) and $sum{bounced}++;
}

sub event_rejected($)
{
	my $t = shift;
	update($t) and $sum{rejected}++;
}

# returns 1 if $sum should be updated
sub update($)
{
	my $t = shift;
	my $m = $t - $t%$rrdstep;
	init_rrd($m) unless $rrd_inited;
	return 1 if $m == $this_minute;
	return 0 if $m < $this_minute;

	#print "$m\n";
	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}\n";
	RRDs::update $rrd, "$this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}";
	if($m > $this_minute+$rrdstep) {
		for(my $sm=$this_minute+$rrdstep;$sm<$m;$sm+=$rrdstep) {
			print "update $sm:0:0:0:0 (SKIP)\n";
			RRDs::update $rrd, "$sm:0:0:0:0";
		}
	}
	$this_minute = $m;
	$sum{sent}=0;
	$sum{received}=0;
	$sum{bounced}=0;
	$sum{rejected}=0;
	return 1;
}

main;
